#!/usr/bin/env groovy

import org.xml.sax.*

final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'

//------------------------------------------------------------------------------
// Utility methods and classes
//------------------------------------------------------------------------------

final def getStepProperties = {
    def props = [:]

    def ahptoolProc = [ahptool, "getStepProperties"].execute()
    ahptoolProc.consumeProcessErrorStream(System.out);
    def propertiesXml = ahptoolProc.text
    ahptoolProc.waitFor()
    if (ahptoolProc.exitValue()) {
        throw new Exception("Failed to get step properties from ahptool: " + propertiesXml)
    }
    new XmlSlurper().parseText(propertiesXml).property.each{ propElem ->
        props[propElem.@name.text()] = propElem.text()
    }

    return props;
}

class CoverityHandler extends helpers.DefaultHandler {

    def findingId
    def findingFile
    def findingLine
    def findingName
    def findingSeverity
    def findingStatus
    def addFinding
    def sb

    CoverityHandler(def addFinding) {
        this.addFinding = addFinding
    }

    void startElement(String ns, String localName, String qName, Attributes attribs) {
        sb = new StringBuilder()
    }

    void characters(char[] ch, int start, int length) {
        sb.append(ch, start, length)
    }

    void endElement(String ns, String localName, String qName) {
        switch (qName) {

            case 'num':
                findingId = sb.toString()
                break

            case 'checker':
                findingName = sb.toString()
                break

            case 'file':
                findingFile = sb.toString()
                break

            case 'line':
                findingLine = sb.toString()
                break

            case 'status':
                findingStatus = sb.toString()
                break

            case 'error':
                // not setting the id because the XML report we are reading is not generating the unique id using in the Defect Manager server
                addFinding('', findingFile, findingLine, findingName, 'ERROR', '', findingStatus)
                break
        }
    }
}

class DummyEntityResolver implements EntityResolver {
    InputSource resolveEntity(String publicID, String systemID)
    throws SAXException {
        return new InputSource(new StringReader(""));
    }
}

final File resourceHome = new File(System.getenv()['PLUGIN_HOME'])

final def workDir = new File('.').canonicalFile
final def out = System.out

final def props = getStepProperties()

final def commandPath = props['integration/commandPath'] ? props['integration/commandPath'] + File.separator : ''
final def covRemote = props['integration/covRemote']
final def covPort = props['integration/covPort']
final def covUsername = props['integration/covUsername']
final def covPassword = props['integration/covPassword']
final def covVersion = props['integration/version']

final def buildLifeId = System.getenv()['AH_BUILD_LIFE_ID']
final def reportName = props['reportName'] ? props['reportName'] : 'Coverity Report'
final def productName = props['productName']
final def streamName = props['streamName']
final def sourceType = props['sourceType']
final def buildCommand = props['buildCommand']
final def directory = '.coverity'

final def buildCommandList = buildCommand.tokenize("\n")
for(String curr : buildCommandList) {
	//white space is breaking things
	curr.trim();
}

try {
    //------------------------------------------------------------------------------
    // Run Coverity Prevent build which creates a intermediate directory
    //------------------------------------------------------------------------------

    def commandLine = []

    commandLine.add(commandPath + 'cov-build')
    commandLine.add('--dir')
    commandLine.add(directory)
    commandLine.addAll(buildCommandList)

    def pb = new ProcessBuilder((String[]) commandLine).directory(workDir)

    println()
    println("command line: ${commandLine.join(' ')}")
    println("working directory: " + workDir)
    println("command environment: ")
    println("command output: ")
    println('-------------------------------')

    // run the process
    def proc = pb.start()
    proc.consumeProcessOutput(out, out);
    proc.waitFor()
    if (proc.exitValue()) {
        println("cov-build run failed with exit code: " + proc.exitValue())
        System.exit(proc.exitValue())
    }
    println('-------------------------------')
    println()


    //------------------------------------------------------------------------------
    // Run Coverity Prevent analyze on the intermediate directory
    //------------------------------------------------------------------------------

    def outputDir = null
    def analyzeCommand = 'cov-analyze'
    commandLine = []

    if ("C#".equalsIgnoreCase(sourceType)) {
	    analyzeCommand = 'cov-analyze-cs'
    }
    else if ("Java".equalsIgnoreCase(sourceType)) {
	    analyzeCommand = 'cov-analyze-java'
    }
    
    commandLine.add(commandPath + analyzeCommand)
    commandLine.add('--dir')
    commandLine.add(directory)

    pb = new ProcessBuilder((String[]) commandLine).directory(workDir)

    println()
    println("command line: ${commandLine.join(' ')}")
    println("working directory: " + workDir)
    println("command environment: ")
    println("command output: ")
    println('-------------------------------')

    // run the process
    proc = pb.start()
    proc.consumeProcessErrorStream(System.out);
    proc.in.eachLine { line ->
        if (line.startsWith('Output goes to:')) {
            outputDir = new File(line.substring(line.indexOf(":") + 1).trim())
        }
        println(line);
    }
    proc.waitFor()
    if (proc.exitValue()) {
        println(analyzeCommand + " run failed with exit code: " + proc.exitValue())
        System.exit(proc.exitValue())
    }
    println('-------------------------------')
    println()
    


    //------------------------------------------------------------------------------
    // Run Coverity Prevent format errors to generate XML and HTML results
    //------------------------------------------------------------------------------

    commandLine = []

    commandLine.add(commandPath + 'cov-format-errors')
    commandLine.add('--dir')
    commandLine.add(directory)

    pb = new ProcessBuilder((String[]) commandLine).directory(workDir)

    println()
    println("command line: ${commandLine.join(' ')}")
    println("working directory: " + workDir)
    println("command environment: ")
    println("command output: ")
    println('-------------------------------')

    // run the process
    proc = pb.start()
    proc.consumeProcessOutput(out, out);
    proc.waitFor()
    if (proc.exitValue()) {
        println("cov-format-errors run failed with exit code: " + proc.exitValue())
        System.exit(proc.exitValue())
    }
    println('-------------------------------')
    println()


    //------------------------------------------------------------------------------
    // Run Coverity Prevent commit defects to send the results to the Defect Manager
    //------------------------------------------------------------------------------

    def runId = null
    commandLine = []

    commandLine.add(commandPath + 'cov-commit-defects')
    commandLine.add('--dir')
    commandLine.add(directory)
    if(covVersion.equals("4")) {
    	commandLine.add('--product')
    	commandLine.add(productName)
    	commandLine.add('--remote')
    	commandLine.add(covRemote)
    }
    else if(covVersion.equals("5")) {
        commandLine.add('--host')
    	commandLine.add(covRemote)
        commandLine.add('--stream') 
        commandLine.add(streamName)
        commandLine.add('--project')
    	commandLine.add(productName)
    }
    commandLine.add('--port')
    commandLine.add(covPort)
    if (covUsername) {
        commandLine.add('--user')
        commandLine.add(covUsername)
    }
    if (covPassword) {
        commandLine.add('--password')
        commandLine.add(covPassword)
    }

    pb = new ProcessBuilder((String[]) commandLine).directory(workDir)

    println()
    println("command line: ${commandLine.join(' ')}")
    println("working directory: " + workDir)
    println("command environment: ")
    println("command output: ")
    println('-------------------------------')

    // run the process
    proc = pb.start()
    proc.consumeProcessErrorStream(System.out);
    if(covVersion.equals("4")) {
        proc.in.eachLine { line ->
            if (line.startsWith('New run ID ')) {
                def scanner = new Scanner(line).useDelimiter('New run ID ');
                scanner = new Scanner(scanner.next()).useDelimiter(' added.');
                runId = scanner.next()
            }
            println(line);
        }
    }
    proc.waitFor()
    if (proc.exitValue()) {
        println("cov-commit-defects run failed with exit code: " + proc.exitValue())
        System.exit(proc.exitValue())
    }
    println('-------------------------------')
    println()


    //------------------------------------------------------------------------------
    // Define a closure for parsing the Coverity XML and writing out Anthill XML format
    //------------------------------------------------------------------------------
    File indexXmlFile = new File(outputDir, 'errors' + File.separator + 'index.xml') 
    def runUrlLink = '';
    if(covVersion.equals("4")) {
        runUrlLink = 'http://' + covRemote + ':' + covPort + '/cov.cgi?run=' + runId + '&t=14&v=1'
    }
    else if(covVersion.equals("5")) {
    	//TODO get projid through webservices
    	runUrlLink = 'http://'+covRemote + ':' + covPort + '/projects/dashboard.htm?projectId=' + projId
    }

    def xml = new groovy.xml.StreamingMarkupBuilder().bind {
        mkp.xmlDeclaration()

        analytics (
            name:reportName, buildLifeId:buildLifeId, type:'Coverity', urlLink: runUrlLink) {

            def addFinding = { findingId, findingFile, findingLine, findingName, findingSeverity, findingDesc, findingStatus ->
                finding() {
                    id(findingId)
                    file(findingFile)
                    line(findingLine)
                    name(findingName)
                    severity(findingSeverity)
                    description(findingDesc)
                    status(findingStatus)
                }
            }

            def handler = new CoverityHandler(addFinding)
            def reader = javax.xml.parsers.SAXParserFactory.newInstance().newSAXParser().XMLReader
            reader.setEntityResolver(new DummyEntityResolver())
            reader.setContentHandler(handler)
            reader.parse(new InputSource(indexXmlFile.newReader()))
        }
    }


    //------------------------------------------------------------------------------
    // Construct XML message and upload to AHP
    //------------------------------------------------------------------------------

    println("Uploading Coverity Prevent results to server as report '" + reportName + "' ....")

    commandLine = [ahptool, '-report', reportName, 'addSourceAnalytics', '-']
    pb = new ProcessBuilder(commandLine).directory(workDir)
    
    println()
    println("command line: ${commandLine.join(' ')}")
    println("working directory: " + workDir)
    println("command environment: ")
    println("command output: ")
    println('-------------------------------')
    
    proc = pb.start()
    proc.consumeProcessOutput(out, out);
    proc.withWriter{ it << xml }
    proc.waitFor()
    if (proc.exitValue()) {
        println("Coverity Prevent results upload to server failed with exit code: " + proc.exitValue())
        System.exit(proc.exitValue())
    }
    else {
        println("Coverity Prevent results uploaded successfully")
    }
    println('-------------------------------')
    println()

}
catch (Exception e) {
    e.printStackTrace()
    System.exit(1)
}